@@ -429,6 +429,8 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        assert_eq!(self.left.shape(), self.right.shape());
+
         debug!(
             "BlockListCombine::transpose {} {} {:?}",
             self.left.shape(),
@@ -438,6 +440,7 @@ where
 
         let left = self.left.transpose(permutation.clone())?;
         let right = self.right.transpose(permutation)?;
+        assert_eq!(left.shape(), right.shape());
 
         BlockListCombine::new(
             left,
@@ -603,6 +606,8 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListConst::transpose {:?}", permutation);
+
         let transpose = self.source.transpose(permutation)?;
         Ok(BlockListConst::new(
             transpose,
@@ -654,6 +659,8 @@ impl<FD, FS, D, T, B> fmt::Display for BlockListConst<FD, FS, D, T, B> {
 pub struct BlockListBroadcast<FD, FS, D, T, B> {
     source: B,
     rebase: transform::Broadcast,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -712,7 +719,7 @@ where
     D::FileClass: From<TensorType>,
     B: DenseAccess<FD, FS, D, T>,
 {
-    type Slice = BlockListBroadcast<FD, FS, D, T, B::Slice>;
+    type Slice = DenseAccessor<FD, FS, D, T>;
     type Transpose = BlockListTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
@@ -755,10 +762,16 @@ where
         let shape = bounds.to_shape(self.shape())?;
         let bounds = self.rebase.invert_bounds(bounds);
         let source = self.source.slice(bounds)?;
-        BlockListBroadcast::new(source, shape)
+
+        if source.shape() == &shape {
+            Ok(source.accessor())
+        } else {
+            BlockListBroadcast::new(source, shape).map(DenseAccess::accessor)
+        }
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListBroadcast::transpose {:?}", permutation);
         BlockListTranspose::new(self, permutation)
     }
 
@@ -801,6 +814,8 @@ impl<FD, FS, D, T, B> fmt::Display for BlockListBroadcast<FD, FS, D, T, B> {
 pub struct BlockListCast<FD, FS, D, T, B> {
     source: B,
     dtype: NumberType,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -881,6 +896,7 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListCast::transpose {:?}", permutation);
         let transpose = self.source.transpose(permutation)?;
         Ok(BlockListCast::new(transpose, self.dtype))
     }
@@ -1032,15 +1048,35 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let expand_axis = if let Some(permutation) = &permutation {
-            permutation[self.rebase.expand_axis()]
-        } else {
-            self.ndim() - self.rebase.expand_axis()
-        };
+        debug!(
+            "BlockListExpand::transpose {} {:?} (expanded axis is {})",
+            self.shape(),
+            permutation,
+            self.rebase.expand_axis(),
+        );
 
-        let permutation = permutation.map(|axes| self.rebase.invert_axes(axes));
-        let source = self.source.transpose(permutation)?;
+        let permutation =
+            permutation.unwrap_or_else(|| (0..self.ndim()).into_iter().rev().collect());
+
+        assert_eq!(permutation.len(), self.ndim());
+
+        let mut expand_axis = None;
+        for i in 0..permutation.len() {
+            if permutation[i] == self.rebase.expand_axis() {
+                expand_axis = Some(i);
+            }
+        }
+        let expand_axis = expand_axis.expect("expand axis");
+
+        let permutation = self.rebase.invert_axes(permutation);
+        let source = self.source.transpose(Some(permutation))?;
+
+        debug!(
+            "BlockListExpand::transpose expand source transpose at axis {}",
+            expand_axis
+        );
         let rebase = transform::Expand::new(source.shape().clone(), expand_axis)?;
+
         Ok(BlockListExpand {
             source,
             rebase,
@@ -1192,6 +1228,8 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListFlip::transpose {:?}", permutation);
+
         let axis = if let Some(permutation) = &permutation {
             if permutation.len() != self.ndim() {
                 return Err(TCError::bad_request(
@@ -1378,6 +1416,8 @@ where
     }
 
     fn block_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
+        debug!("BlockListReduce::block_stream with shape {}", self.shape());
+
         Box::pin(async move {
             let reductor = self.reductor;
             let axis = self.rebase.reduce_axis();
@@ -1389,10 +1429,20 @@ where
                 Ok(reductor.reduce_stream(blocks))
             } else {
                 let mut permutation: Vec<usize> = (0..ndim).collect();
-                permutation[axis] = ndim - 1;
-                permutation[ndim - 1] = axis;
+                let axis = permutation.remove(axis);
+                permutation.push(axis);
+
+                debug!(
+                    "BlockListReduce::block_stream will transpose its source: {:?}",
+                    permutation
+                );
 
                 let transpose = source.transpose(Some(permutation))?;
+                debug!(
+                    "reducing last axis of dense block list with shape {}",
+                    transpose.shape()
+                );
+
                 let blocks = transpose.block_stream(txn).await?;
                 Ok(reductor.reduce_stream(blocks))
             }
@@ -1419,7 +1469,9 @@ where
             permutation
         );
 
-        BlockListTranspose::new(self, permutation)
+        let transpose = BlockListTranspose::new(self, permutation)?;
+        debug!("BlockListReduce::transpose shape is {}", transpose.shape());
+        Ok(transpose)
     }
 
     async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
@@ -1476,6 +1528,8 @@ impl<FD, FS, D, T, B> fmt::Display for BlockListReduce<FD, FS, D, T, B> {
 pub struct BlockListReshape<FD, FS, D, T, B> {
     source: B,
     rebase: transform::Reshape,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -1558,12 +1612,14 @@ where
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
         debug!(
-            "BlockListReduce::transpose {} {:?}",
+            "BlockListReshape::transpose {} {:?}",
             self.shape(),
             permutation
         );
 
-        BlockListTranspose::new(self, permutation)
+        let transpose = BlockListTranspose::new(self, permutation)?;
+        debug!("BlockListReshape::transpose shape is {}", transpose.shape());
+        Ok(transpose)
     }
 
     async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
@@ -1704,6 +1760,8 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListTranspose::transpose {:?}", permutation);
+
         let permutation = if let Some(permutation) = permutation {
             self.rebase.invert_axes(permutation)
         } else {
@@ -1792,7 +1850,7 @@ where
     A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = BlockListSparse<FD, FS, D, T, A::Slice>;
-    type Transpose = BlockListSparse<FD, FS, D, T, A::Transpose>;
+    type Transpose = BlockListSparse<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>;
 
     fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
         let source = self.source.accessor();
@@ -1822,6 +1880,8 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListSparse::transpose {:?}", permutation);
+
         let transpose = self.source.transpose(permutation)?;
         Ok(transpose.into())
     }
@@ -1878,6 +1938,8 @@ pub struct BlockListUnary<FD, FS, D, T, B> {
     transform: fn(&Array) -> Array,
     value_transform: fn(Number) -> Number,
     dtype: NumberType,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -1978,6 +2040,8 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("BlockListUnary::transpose {:?}", permutation);
+
         let source = self.source.transpose(permutation)?;
         Ok(BlockListUnary {
             source,
