@@ -47,6 +47,8 @@ pub const PER_BLOCK: usize = 131_072;
 #[derive(Clone)]
 pub struct DenseTensor<FD, FS, D, T, B> {
     blocks: B,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -327,6 +329,7 @@ where
     B: DenseAccess<FD, FS, D, T>,
 {
     type Combine = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
+    type DenseCombine = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
 
     fn and_const(self, other: Number) -> TCResult<Self::Combine> {
         fn array_and(l: Array, r: Number) -> Array {
@@ -941,6 +944,7 @@ where
     B: DenseAccess<FD, FS, D, T>,
 {
     type Combine = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
+    type DenseCombine = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
 
     fn add_const(self, other: Number) -> TCResult<Self::Combine> {
         fn add_array(l: Array, r: Number) -> Array {
@@ -1090,7 +1094,7 @@ where
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    type Broadcast = DenseTensor<FD, FS, D, T, BlockListBroadcast<FD, FS, D, T, B>>;
+    type Broadcast = DenseTensor<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>;
     type Cast = DenseTensor<FD, FS, D, T, BlockListCast<FD, FS, D, T, B>>;
     type Expand = DenseTensor<FD, FS, D, T, BlockListExpand<FD, FS, D, T, B>>;
     type Flip = DenseTensor<FD, FS, D, T, BlockListFlip<FD, FS, D, T, B>>;
@@ -1099,8 +1103,12 @@ where
     type Transpose = DenseTensor<FD, FS, D, T, B::Transpose>;
 
     fn broadcast(self, shape: Shape) -> TCResult<Self::Broadcast> {
+        if self.shape() == &shape {
+            return Ok(self.into_inner().accessor().into());
+        }
+
         let blocks = BlockListBroadcast::new(self.blocks, shape)?;
-        Ok(DenseTensor::from(blocks))
+        Ok(DenseTensor::from(blocks.accessor()))
     }
 
     fn cast_into(self, dtype: NumberType) -> TCResult<Self::Cast> {
