@@ -14,7 +14,7 @@ use tc_btree::{BTreeType, Node};
 use tc_error::*;
 use tc_transact::fs::{Dir, File};
 use tc_transact::{IntoView, Transaction, TxnId};
-use tc_value::{FloatType, IntType, Number, NumberType, UIntType, Value, ValueType};
+use tc_value::{ComplexType, FloatType, IntType, Number, NumberType, UIntType, Value, ValueType};
 use tcgeneric::{
     label, path_label, Class, Instance, NativeClass, PathLabel, PathSegment, TCBoxTryFuture,
     TCPathBuf, Tuple,
@@ -54,7 +54,22 @@ impl Schema {
     ///
     /// `debug_info` should be the name of the calling Op, as a user would understand it.
     pub fn validate(&self, debug_info: &'static str) -> TCResult<()> {
-        self.shape.validate(debug_info)
+        self.shape.validate(debug_info)?;
+
+        fn err_nonspecific(dtype: NumberType) -> TCResult<()> {
+            Err(TCError::bad_request(
+                "tensor requires a specific Number type, not",
+                dtype,
+            ))
+        }
+
+        match self.dtype {
+            NumberType::Complex(ComplexType::Complex) => err_nonspecific(self.dtype),
+            NumberType::Float(FloatType::Float) => err_nonspecific(self.dtype),
+            NumberType::Int(IntType::Int) => err_nonspecific(self.dtype),
+            NumberType::UInt(UIntType::UInt) => err_nonspecific(self.dtype),
+            _ => Ok(()),
+        }
     }
 }
 
@@ -170,17 +185,19 @@ pub trait TensorBoolean<O> {
 
 /// [`Tensor`] boolean operations in relation to a constant.
 pub trait TensorBooleanConst {
-    /// The result type of a boolean operation.
+    /// The return type of a boolean operation.
     type Combine: TensorInstance;
+    /// The return type of a boolean operation with a result expected to be dense.
+    type DenseCombine: TensorInstance;
 
     /// Logical and
     fn and_const(self, other: Number) -> TCResult<Self::Combine>;
 
     /// Logical or
-    fn or_const(self, other: Number) -> TCResult<Self::Combine>;
+    fn or_const(self, other: Number) -> TCResult<Self::DenseCombine>;
 
     /// Logical xor
-    fn xor_const(self, other: Number) -> TCResult<Self::Combine>;
+    fn xor_const(self, other: Number) -> TCResult<Self::DenseCombine>;
 }
 
 /// Tensor comparison operations
@@ -317,11 +334,13 @@ pub trait TensorMath<D: Dir, O> {
 
 /// [`Tensor`] constant math operations
 pub trait TensorMathConst {
-    /// The result type of a math operation
+    /// The return type of a math operation
     type Combine: TensorInstance;
+    /// The return type of a math operation with a result expected to be dense
+    type DenseCombine: TensorInstance;
 
     /// Add a constant to this tensor
-    fn add_const(self, other: Number) -> TCResult<Self::Combine>;
+    fn add_const(self, other: Number) -> TCResult<Self::DenseCombine>;
 
     /// Divide `self` by `other`.
     fn div_const(self, other: Number) -> TCResult<Self::Combine>;
@@ -336,7 +355,7 @@ pub trait TensorMathConst {
     fn pow_const(self, other: Number) -> TCResult<Self::Combine>;
 
     /// Subtract `other` from `self`.
-    fn sub_const(self, other: Number) -> TCResult<Self::Combine>;
+    fn sub_const(self, other: Number) -> TCResult<Self::DenseCombine>;
 }
 
 /// Methods to access this [`Tensor`] as a persistent type.
@@ -684,6 +703,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Self;
+    type DenseCombine = Self;
 
     fn and_const(self, other: Number) -> TCResult<Self::Combine> {
         match self {
@@ -989,6 +1009,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Self;
+    type DenseCombine = Self;
 
     fn add_const(self, other: Number) -> TCResult<Self::Combine> {
         match self {
@@ -1148,10 +1169,16 @@ where
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self> {
         if let Some(permutation) = &permutation {
             if permutation.iter().zip(0..self.ndim()).all(|(x, i)| x == &i) {
+                debug!(
+                    "no need to transpose {} with permutation {:?}",
+                    self, permutation
+                );
                 return Ok(self);
             }
         }
 
+        debug!("transpose {} with permutation {:?}", self, permutation);
+
         match self {
             Self::Dense(dense) => dense.transpose(permutation).map(Self::from),
             Self::Sparse(sparse) => sparse.transpose(permutation).map(Self::from),
@@ -1552,3 +1579,12 @@ fn coord_bounds(shape: &[u64]) -> Vec<u64> {
         .map(|axis| shape[axis + 1..].iter().product())
         .collect()
 }
+
+#[inline]
+fn needs_transpose(permutation: &Option<Vec<usize>>) -> bool {
+    if let Some(axes) = permutation {
+        !axes.iter().enumerate().all(|(i, x)| x == &i)
+    } else {
+        true
+    }
+}
