@@ -1,11 +1,11 @@
 use std::fmt;
 
-use afarray::{Array, CoordBlocks, CoordMerge, Coords};
+use afarray::{Array, CoordBlocks, CoordMerge, CoordUnique, Coords};
 use async_trait::async_trait;
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, StreamExt, TryStreamExt};
 use futures::try_join;
-use log::debug;
+use log::{debug, trace};
 use safecast::AsType;
 
 use tc_btree::Node;
@@ -18,8 +18,8 @@ use tcgeneric::{TCBoxTryFuture, TCBoxTryStream, Tuple};
 use crate::dense::{DenseAccess, DenseAccessor, DenseTensor, PER_BLOCK};
 use crate::stream::{sorted_coords, sorted_values, Read, ReadValueAt};
 use crate::{
-    coord_bounds, transform, AxisBounds, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType,
-    TensorUnary, ERR_INF, ERR_NAN,
+    coord_bounds, needs_transpose, transform, AxisBounds, Bounds, Coord, Phantom, Shape,
+    TensorAccess, TensorType, TensorUnary, ERR_INF, ERR_NAN,
 };
 
 use super::combine::{coord_to_offset, SparseCombine};
@@ -34,9 +34,6 @@ pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D
     /// The type of a slice of this accessor
     type Slice: SparseAccess<FD, FS, D, T>;
 
-    /// The type of a transpose of this accessor
-    type Transpose: SparseAccess<FD, FS, D, T>;
-
     /// Return this accessor as a [`SparseAccessor`].
     fn accessor(self) -> SparseAccessor<FD, FS, D, T>;
 
@@ -54,13 +51,16 @@ pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D
     /// Return the number of nonzero values in this [`SparseTensor`].
     async fn filled_count(self, txn: T) -> TCResult<u64>;
 
+    /// Return `true` if this [`SparseTensor`] contains no elements.
+    async fn is_empty(self, txn: T) -> TCResult<bool>;
+
     /// Return a slice of this accessor with the given [`Bounds`].
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice>;
 
     /// Return this accessor as transposed according to the given `permutation`.
     ///
     /// If no permutation is given, this accessor's axes will be reversed.
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose>;
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>>;
 }
 
 /// Write methods for [`SparseTensor`] data
@@ -149,7 +149,6 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = Self;
-    type Transpose = Self;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         self
@@ -167,6 +166,10 @@ where
         dispatch!(self, this, this.filled_count(txn).await)
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        dispatch!(self, this, this.is_empty(txn).await)
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self> {
         dispatch!(self, this, this.slice(bounds).map(SparseAccess::accessor))
     }
@@ -263,7 +266,6 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = DenseToSparse<FD, FS, D, T, B::Slice>;
-    type Transpose = DenseToSparse<FD, FS, D, T, B::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Dense(Box::new(DenseToSparse {
@@ -286,6 +288,8 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("DenseToSparse::filled_at {:?}", axes);
+
         if axes.is_empty() {
             return Ok(Box::pin(stream::empty()));
         }
@@ -340,12 +344,28 @@ where
             .await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        DenseTensor::from(self.source)
+            .any(txn)
+            .map_ok(|any| !any)
+            .await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         self.source.slice(bounds).map(DenseToSparse::from)
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        self.source.transpose(permutation).map(DenseToSparse::from)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("DenseToSparse::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            self.source
+                .transpose(permutation)
+                .map(DenseToSparse::from)
+                .map(SparseAccess::accessor)
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -383,6 +403,8 @@ impl<FD, FS, D, T, B> From<B> for DenseToSparse<FD, FS, D, T, B> {
 pub struct SparseBroadcast<FD, FS, D, T, A> {
     source: A,
     rebase: transform::Broadcast,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -398,7 +420,9 @@ where
 {
     pub fn new(source: A, shape: Shape) -> TCResult<Self> {
         debug!("SparseBroadcast::new {} into {}", source.shape(), shape);
+
         let rebase = transform::Broadcast::new(source.shape().clone(), shape)?;
+
         Ok(Self {
             source,
             rebase,
@@ -443,8 +467,7 @@ where
     D::FileClass: From<TensorType>,
     A: SparseAccess<FD, FS, D, T>,
 {
-    type Slice = SparseBroadcast<FD, FS, D, T, A::Slice>;
-    type Transpose = SparseTranspose<FD, FS, D, T, Self>;
+    type Slice = SparseAccessor<FD, FS, D, T>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Broadcast(Box::new(SparseBroadcast {
@@ -455,7 +478,7 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
-        debug!("SparseBroadcast::filled");
+        debug!("SparseBroadcast::filled with source {}", self.source);
 
         let source_axes: Vec<usize> = (0..self.source.ndim()).collect();
         let ndim = self.ndim();
@@ -476,20 +499,25 @@ where
             .try_flatten();
 
         let coords = CoordBlocks::new(coords, ndim, PER_BLOCK);
+
         let broadcast = sorted_values::<FD, FS, T, D, _, _>(txn, self, coords).await?;
         Ok(Box::pin(broadcast))
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
         debug!(
-            "SparseBroadcast::filled_at {:?} (source is {})",
-            axes, self.source
+            "SparseBroadcast::filled_at {:?} (source is {}, shape is {})",
+            axes,
+            self.source,
+            self.rebase.shape()
         );
 
         if axes.is_empty() {
             return Ok(Box::pin(stream::empty()));
         }
 
+        // TODO: optimize for the case when none of the `axes` have been broadcast
+
         self.shape().validate_axes(&axes)?;
 
         let shape = Shape::from({
@@ -497,7 +525,7 @@ where
             axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>()
         });
 
-        let ndim = self.ndim();
+        let ndim = axes.len();
         let rebase = self.rebase;
         let source_axes = (0..self.source.ndim()).collect();
 
@@ -506,15 +534,24 @@ where
             .map_ok(|coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
             .map_ok(move |coord| {
-                debug!("broadcast source coord {:?}", coord);
-                rebase.map_coord(coord)
+                trace!("broadcast source coord {:?}", coord);
+
+                let bounds = rebase.map_coord(coord);
+                let bounds: Bounds = bounds
+                    .into_iter()
+                    .enumerate()
+                    .filter_map(|(x, bound)| if axes.contains(&x) { Some(bound) } else { None })
+                    .collect();
+
+                assert_eq!(bounds.len(), ndim);
+                bounds
             })
             // TODO: can this happen in `Coords`? maybe a new stream generator type?
+            .inspect_ok(|bounds: &Bounds| trace!("broadcast coord mapped to bounds {}", bounds))
             .map_ok(move |bounds| stream::iter(bounds.affected().map(TCResult::Ok)))
             .try_flatten();
 
-        let filled_at =
-            CoordBlocks::new(filled_at, ndim, PER_BLOCK).map_ok(move |coords| coords.get(&axes));
+        let filled_at = CoordBlocks::new(filled_at, ndim, PER_BLOCK);
 
         let filled_at = sorted_coords::<FD, FS, D, T, _>(&txn, shape, filled_at).await?;
 
@@ -532,6 +569,10 @@ where
             .await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseBroadcast::slice {} from {}", bounds, self.shape());
         self.shape().validate_bounds(&bounds)?;
@@ -546,11 +587,22 @@ where
         );
 
         let source = self.source.slice(source_bounds)?;
-        SparseBroadcast::new(source, shape)
+
+        if source.shape() == &shape {
+            Ok(source.accessor())
+        } else {
+            SparseBroadcast::new(source, shape).map(SparseAccess::accessor)
+        }
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        SparseTranspose::new(self, permutation)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseBroadcast::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            SparseTranspose::new(self, permutation).map(SparseAccess::accessor)
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -633,11 +685,12 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D: Dir,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     T: Transaction<D>,
     A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseCast<FD, FS, D, T, A::Slice>;
-    type Transpose = SparseCast<FD, FS, D, T, A::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Cast(Box::new(SparseCast::new(
@@ -655,6 +708,8 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseCast::filled_at {:?}", axes);
+
         self.source.filled_at(txn, axes).await
     }
 
@@ -662,6 +717,10 @@ where
         self.source.filled_count(txn).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         Ok(SparseCast {
             source: self.source.slice(bounds)?,
@@ -670,13 +729,21 @@ where
         })
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let source = self.source.transpose(permutation)?;
-        Ok(SparseCast {
-            source,
-            dtype: self.dtype,
-            phantom: self.phantom,
-        })
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseCast::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            let source = self.source.transpose(permutation)?;
+            let cast = SparseCast {
+                source,
+                dtype: self.dtype,
+                phantom: self.phantom,
+            };
+
+            Ok(cast.accessor())
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -811,11 +878,12 @@ where
     FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     L: SparseAccess<FD, FS, D, T>,
     R: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseCombinator<FD, FS, D, T, L::Slice, R::Slice>;
-    type Transpose = SparseCombinator<FD, FS, D, T, L::Transpose, R::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Combine(Box::new(SparseCombinator {
@@ -829,12 +897,18 @@ where
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
         let zero = self.dtype().zero();
         let filled_inner = self.filled_inner(txn).await?;
-        let filled = filled_inner.try_filter(move |(_, value)| future::ready(*value != zero));
+        let filled = filled_inner
+            .try_filter(move |(_, value)| future::ready(*value != zero))
+            .inspect_ok(|(coord, value)| {
+                trace!("SparseCombinator filled at {:?}: {}", coord, value)
+            });
 
         Ok(Box::pin(filled))
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseCombinator::filled_at {:?}", axes);
+
         self.shape().validate_axes(&axes)?;
 
         if axes.is_empty() {
@@ -863,6 +937,11 @@ where
             .await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        let mut filled = self.filled(txn).await?;
+        filled.try_next().map_ok(|row| row.is_none()).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseCombinator::slice {}", bounds);
 
@@ -878,17 +957,25 @@ where
         })
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseCombinator::transpose {:?}", permutation);
+
+        if !needs_transpose(&permutation) {
+            return Ok(self.accessor());
+        }
+
         let left = self.left.transpose(permutation.clone())?;
         let right = self.right.transpose(permutation)?;
         assert_eq!(left.shape(), right.shape());
 
-        Ok(SparseCombinator {
+        let transpose = SparseCombinator {
             left,
             right,
             combinator: self.combinator,
             phantom: self.phantom,
-        })
+        };
+
+        Ok(transpose.accessor())
     }
 }
 
@@ -970,11 +1057,12 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D: Dir,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     T: Transaction<D>,
     A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseConstCombinator<FD, FS, D, T, A::Slice>;
-    type Transpose = SparseConstCombinator<FD, FS, D, T, A::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         let this = SparseConstCombinator {
@@ -988,15 +1076,19 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        let zero = self.dtype().zero();
         let combinator = self.combinator;
         let other = self.other;
         let filled = self.source.filled(txn).await?;
-        Ok(Box::pin(filled.map_ok(move |(coord, value)| {
-            (coord, combinator(value, other))
-        })))
+        let filled = filled
+            .map_ok(move |(coord, value)| (coord, combinator(value, other)))
+            .try_filter(move |(_coord, value)| future::ready(value != &zero));
+
+        Ok(Box::pin(filled))
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseConstCombinator::filled_at {:?}", axes);
         self.source.filled_at(txn, axes).await
     }
 
@@ -1004,6 +1096,11 @@ where
         self.source.filled_count(txn).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        let mut filled = self.filled(txn).await?;
+        filled.try_next().map_ok(|row| row.is_none()).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         let slice = self.source.slice(bounds)?;
         Ok(SparseConstCombinator::new(
@@ -1013,13 +1110,17 @@ where
         ))
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let transpose = self.source.transpose(permutation)?;
-        Ok(SparseConstCombinator::new(
-            transpose,
-            self.other,
-            self.combinator,
-        ))
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseConstCombinator::transpose {:?}", permutation);
+
+        if !needs_transpose(&permutation) {
+            return Ok(self.accessor());
+        }
+
+        let source = self.source.transpose(permutation)?;
+        let transpose = SparseConstCombinator::new(source, self.other, self.combinator);
+
+        Ok(transpose.accessor())
     }
 }
 
@@ -1102,6 +1203,9 @@ where
                     .read_value_at(txn.clone(), coord)
                     .map_ok(move |(coord, right_value)| (coord, left_value, right_value))
             })
+            .inspect_ok(|(coord, left, right)| {
+                trace!("left combine {} with {} at {:?}", left, right, coord)
+            })
             .map_ok(move |(coord, left, right)| (coord, combinator(left, right)));
 
         Ok(Box::pin(filled))
@@ -1141,11 +1245,12 @@ where
     FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     L: SparseAccess<FD, FS, D, T>,
     R: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseLeftCombinator<FD, FS, D, T, L::Slice, R::Slice>;
-    type Transpose = SparseLeftCombinator<FD, FS, D, T, L::Transpose, R::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::CombineLeft(Box::new(SparseLeftCombinator {
@@ -1159,12 +1264,30 @@ where
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
         let zero = self.dtype().zero();
         let filled_inner = self.filled_inner(txn).await?;
-        let filled = filled_inner.try_filter(move |(_, value)| future::ready(value != &zero));
+        let filled = filled_inner
+            .try_filter(move |(_, value)| future::ready(value != &zero))
+            .inspect_ok(|(coord, value)| {
+                trace!("SparseLeftCombinator filled at {:?}: {}", coord, value)
+            });
+
         Ok(Box::pin(filled))
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
-        self.left.filled_at(txn, axes).await
+        let ndim = self.ndim();
+        let shape: Shape = axes.iter().map(|x| self.shape()[*x]).collect();
+        let filled = self.filled(txn.clone()).await?;
+        let filled = filled.map_ok(|(coord, _value)| coord);
+
+        let coords = CoordBlocks::new(filled, ndim, PER_BLOCK);
+        let filled_at = coords.map_ok(move |coords| coords.get(&axes));
+        let unique = sorted_coords::<FD, FS, D, T, _>(&txn, shape.clone(), filled_at).await?;
+
+        Ok(Box::pin(CoordUnique::new(
+            unique,
+            shape.to_vec(),
+            PER_BLOCK,
+        )))
     }
 
     async fn filled_count(self, txn: T) -> TCResult<u64> {
@@ -1175,6 +1298,11 @@ where
             .await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        let mut filled = self.filled(txn).await?;
+        filled.try_next().map_ok(|row| row.is_none()).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseLeftCombinator::slice left {} {}", self.left, bounds);
         let left = self.left.slice(bounds.clone())?;
@@ -1183,6 +1311,7 @@ where
             "SparseLeftCombinator::slice right {} {}",
             self.right, bounds
         );
+
         let right = self.right.slice(bounds)?;
 
         assert_eq!(left.shape(), right.shape());
@@ -1195,17 +1324,25 @@ where
         })
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseLeftCombinator::transpose {:?}", permutation);
+
+        if !needs_transpose(&permutation) {
+            return Ok(self.accessor());
+        }
+
         let left = self.left.transpose(permutation.clone())?;
         let right = self.right.transpose(permutation)?;
         assert_eq!(left.shape(), right.shape());
 
-        Ok(SparseLeftCombinator {
+        let transpose = SparseLeftCombinator {
             left,
             right,
             combinator: self.combinator,
             phantom: self.phantom,
-        })
+        };
+
+        Ok(transpose.accessor())
     }
 }
 
@@ -1249,6 +1386,8 @@ impl<FD, FS, D, T, L, R> fmt::Display for SparseLeftCombinator<FD, FS, D, T, L,
 pub struct SparseExpand<FD, FS, D, T, A> {
     source: A,
     rebase: transform::Expand,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -1307,7 +1446,6 @@ where
     A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
-    type Transpose = SparseExpand<FD, FS, D, T, A::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Expand(Box::new(SparseExpand {
@@ -1318,9 +1456,14 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!("SparseExpand::filled");
+
         let rebase = self.rebase;
         let filled = self.source.filled(txn).await?;
-        let filled = filled.map_ok(move |(coord, value)| (rebase.map_coord(coord), value));
+        let filled = filled
+            .map_ok(move |(coord, value)| (rebase.map_coord(coord), value))
+            .inspect_ok(|(coord, value)| trace!("SparseExpand::filled at {:?}: {}", coord, value));
+
         Ok(Box::pin(filled))
     }
 
@@ -1349,7 +1492,27 @@ where
             axes, expand
         );
 
+        let mut filled = self.source.clone().filled(txn.clone()).await?;
+        while let Some((coord, value)) = filled.try_next().await? {
+            trace!("source filled at {:?}: {}", coord, value)
+        }
+
         let source_axes = self.rebase.invert_axes(axes);
+        if source_axes.is_empty() {
+            return if self.source.is_empty(txn).await? {
+                Ok(Box::pin(stream::empty()))
+            } else {
+                // the expanded axis has a size of one, so only coordinate 0 can possibly be filled
+                let filled_at = Coords::from_iter([vec![0]], 1);
+                Ok(Box::pin(stream::once(future::ready(Ok(filled_at)))))
+            };
+        }
+
+        debug!(
+            "SparseExpand::filled_at will expand axis {:?} of source {} axes {:?}",
+            expand, self.source, source_axes
+        );
+
         let source = self.source.filled_at(txn, source_axes).await?;
         if let Some(x) = expand {
             let filled_at = source.map_ok(move |coords| coords.expand_dim(x));
@@ -1363,69 +1526,58 @@ where
         self.source.filled_count(txn).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
-        debug!("SparseExpand {} slice {}", self.shape(), bounds);
-        self.shape().validate_bounds(&bounds)?;
-        bounds.normalize(self.shape());
-
-        if bounds == Bounds::all(self.shape()) {
-            return Ok(self.accessor());
-        }
-
-        let source_bounds = self.rebase.invert_bounds(bounds.clone());
-        let source_slice_shape = source_bounds.to_shape(self.source.shape())?;
-
-        let expand_axis = self.rebase.expand_axis();
-        let source_expand_axis = if bounds[expand_axis].is_index() {
-            // in this case the expanded dimension is elided
-            None
-        } else if source_slice_shape.len() == 1 {
-            // in this case only the expanded dimension is present in the slice
-            None
-        } else {
-            let num_elided = bounds[..expand_axis]
-                .iter()
-                .filter(|bound| bound.is_index())
-                .count();
-
-            Some(expand_axis - num_elided)
-        };
-
         debug!(
-            "SparseExpand slice source {} with bounds {} (removed axis {}, expanding axis {:?})",
-            self.source.shape(),
-            source_bounds,
-            expand_axis,
-            source_expand_axis,
+            "SparseExpand {} axis {} slice {}",
+            self.shape(),
+            self.rebase.expand_axis(),
+            bounds
         );
 
-        let slice = self.source.slice(source_bounds)?;
-        debug_assert_eq!(&source_slice_shape, slice.shape());
+        self.shape().validate_bounds(&bounds)?;
+        bounds.normalize(self.shape());
+        let ndim = bounds.ndim();
 
-        if let Some(source_expand_axis) = source_expand_axis {
-            SparseExpand::new(slice, source_expand_axis).map(|expansion| expansion.accessor())
+        let expand_axis = self.rebase.invert_axis(&bounds);
+        let bounds = self.rebase.invert_bounds(bounds);
+        let source = self.source.slice(bounds)?;
+
+        if ndim == source.ndim() {
+            Ok(source.accessor())
+        } else if let Some(axis) = expand_axis {
+            SparseExpand::new(source, axis).map(SparseAccess::accessor)
         } else {
-            Ok(slice.accessor())
+            Ok(source.accessor())
         }
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let expand_axis = if let Some(permutation) = &permutation {
-            if permutation.len() != self.ndim() {
-                return Err(TCError::unsupported(format!(
-                    "invalid permutation for tensor of shape {}: {:?}",
-                    self.shape(),
-                    permutation
-                )));
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseExpand::transpose {:?}", permutation);
+
+        if !needs_transpose(&permutation) {
+            return Ok(self.accessor());
+        }
+
+        let permutation =
+            permutation.unwrap_or_else(|| (0..self.ndim()).into_iter().rev().collect());
+
+        assert_eq!(permutation.len(), self.ndim());
+
+        let mut expand_axis = None;
+        for i in 0..permutation.len() {
+            if permutation[i] == self.rebase.expand_axis() {
+                expand_axis = Some(i);
             }
+        }
+        let expand_axis = expand_axis.expect("expand axis");
 
-            permutation[self.rebase.expand_axis()]
-        } else {
-            self.ndim() - self.rebase.expand_axis()
-        };
-
-        let source = self.source.transpose(permutation)?;
-        SparseExpand::new(source, expand_axis)
+        let permutation = self.rebase.invert_axes(permutation);
+        let source = self.source.transpose(Some(permutation))?;
+        SparseExpand::new(source, expand_axis).map(SparseAccess::accessor)
     }
 }
 
@@ -1460,6 +1612,8 @@ impl<FD, FS, D, T, A> fmt::Display for SparseExpand<FD, FS, D, T, A> {
 pub struct SparseFlip<FD, FS, D, T, A> {
     source: A,
     rebase: transform::Flip,
+
+    #[allow(dead_code)]
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -1519,7 +1673,6 @@ where
     A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
-    type Transpose = SparseFlip<FD, FS, D, T, A::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Flip(Box::new(SparseFlip {
@@ -1558,6 +1711,10 @@ where
         self.source.filled_count(txn).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseFlip {} slice {}", self.shape(), bounds);
 
@@ -1580,7 +1737,13 @@ where
         }
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseFlip::transpose {:?}", permutation);
+
+        if !needs_transpose(&permutation) {
+            return Ok(self.accessor());
+        }
+
         let flip_axis = if let Some(permutation) = &permutation {
             if permutation.len() == self.ndim() {
                 permutation[self.rebase.axis()]
@@ -1595,7 +1758,7 @@ where
         };
 
         let source = self.source.transpose(permutation)?;
-        SparseFlip::new(source, flip_axis)
+        SparseFlip::new(source, flip_axis).map(SparseAccess::accessor)
     }
 }
 
@@ -1695,7 +1858,6 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = SparseReduce<FD, FS, D, T>;
-    type Transpose = SparseTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Reduce(Box::new(self))
@@ -1705,10 +1867,16 @@ where
         debug!("SparseReduce::filled");
 
         let filled_at = self
+            .source
             .clone()
-            .filled_at(txn.clone(), (0..self.ndim()).collect())
+            .filled_at(
+                txn.clone(),
+                self.rebase
+                    .invert_axes((0..self.ndim()).into_iter().collect()),
+            )
             .await?;
 
+        let ndim = self.ndim();
         let zero = self.dtype().zero();
         let rebase = self.rebase;
         let reductor = self.reductor;
@@ -1718,8 +1886,11 @@ where
             .map_ok(|coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
             .map_ok(move |coord| {
+                debug_assert_eq!(coord.len(), ndim);
+
                 let source_bounds = rebase.invert_coord(&coord);
-                debug!("reduce {:?} (bounds {})", coord, source_bounds);
+                trace!("reduce {:?} (bounds {})", coord, source_bounds);
+
                 let source = source.clone();
                 let txn = txn.clone();
                 Box::pin(async move {
@@ -1729,16 +1900,26 @@ where
                 })
             })
             .try_buffered(num_cpus::get())
-            .try_filter(move |(_coord, value)| future::ready(value != &zero));
+            .try_filter(move |(_coord, value)| future::ready(value != &zero))
+            .inspect_ok(|(coord, value)| trace!("SparseReduce filled at {:?}: {}", coord, value));
 
         Ok(Box::pin(filled))
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
-        debug!("SparseReduce::filled_at {:?}", axes);
+        let ndim = self.ndim();
+        let shape: Vec<u64> = axes.iter().map(|x| self.shape()[*x]).collect();
 
-        let source_axes = self.rebase.invert_axes(axes);
-        self.source.filled_at(txn, source_axes).await
+        let filled = self.filled(txn.clone()).await?;
+        let filled = filled.map_ok(|(coord, _value)| coord);
+        let filled =
+            CoordBlocks::new(filled, ndim, PER_BLOCK).map_ok(move |coords| coords.get(&axes));
+
+        let sorted = sorted_coords::<FD, FS, D, T, _>(&txn, shape.to_vec().into(), filled).await?;
+
+        let unique = CoordUnique::new(sorted, shape, PER_BLOCK);
+
+        Ok(Box::pin(unique))
     }
 
     async fn filled_count(self, txn: T) -> TCResult<u64> {
@@ -1750,8 +1931,13 @@ where
             .await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        let mut filled = self.filled(txn).await?;
+        filled.try_next().map_ok(|row| row.is_none()).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
-        debug!("SparseReduce::slice {}", bounds);
+        debug!("SparseReduce::slice {} from {:?}", bounds, self.shape());
 
         self.shape().validate_bounds(&bounds)?;
 
@@ -1764,6 +1950,8 @@ where
 
         let source = self.source.slice(source_bounds)?;
         let reduced = SparseReduce::new(source.into(), reduce_axis, self.reductor)?;
+        debug_assert!(reduced.ndim() > 0);
+
         if reduced.ndim() == 0 {
             Err(TCError::unsupported(
                 "cannot return a zero-dimensional slice from a reduced Tensor",
@@ -1773,8 +1961,14 @@ where
         }
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        SparseTranspose::new(self, permutation)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseReduce::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            SparseTranspose::new(self, permutation).map(SparseAccess::accessor)
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -1868,7 +2062,6 @@ where
     A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseTable<FD, FS, D, T>;
-    type Transpose = SparseTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         let reshape = SparseReshape {
@@ -1909,14 +2102,24 @@ where
             .await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, _bounds: Bounds) -> TCResult<Self::Slice> {
         Err(TCError::unsupported(
             "cannot slice a reshaped Tensor; make a copy first",
         ))
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        SparseTranspose::new(self, permutation)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseReshape::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            SparseTranspose::new(self, permutation).map(SparseAccess::accessor)
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -1958,6 +2161,16 @@ where
     A: SparseAccess<FD, FS, D, T>,
 {
     pub fn new(source: A, permutation: Option<Vec<usize>>) -> TCResult<Self> {
+        if !needs_transpose(&permutation) {
+            return Err(TCError::bad_request(
+                "sparse transpose got no axes to transpose",
+                permutation
+                    .expect("permutation")
+                    .iter()
+                    .collect::<Tuple<&usize>>(),
+            ));
+        }
+
         transform::Transpose::new(source.shape().clone(), permutation).map(|rebase| Self {
             source,
             rebase,
@@ -2003,8 +2216,7 @@ where
     A: SparseAccess<FD, FS, D, T>,
     A::Slice: SparseAccess<FD, FS, D, T>,
 {
-    type Slice = <A::Slice as SparseAccess<FD, FS, D, T>>::Transpose;
-    type Transpose = A::Transpose;
+    type Slice = SparseAccessor<FD, FS, D, T>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Transpose(Box::new(SparseTranspose {
@@ -2015,8 +2227,10 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!("SparseTranspose::filled with source {}", self.source);
+
         let rebase = self.rebase.clone();
-        let source_axes = self.rebase.invert_axes((0..self.ndim()).collect());
+        let source_axes = (0..self.ndim()).collect();
         let filled_at = self
             .source
             .clone()
@@ -2054,6 +2268,10 @@ where
         self.source.filled_count(txn).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseTranspose::slice {}", bounds);
         self.shape().validate_bounds(&bounds)?;
@@ -2071,9 +2289,22 @@ where
         source.transpose(Some(slice_permutation))
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let permutation = permutation.map(|axes| self.rebase.invert_axes(axes));
-        self.source.transpose(permutation)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseTranspose::transpose {:?}", permutation);
+
+        if !needs_transpose(&permutation) {
+            return Ok(self.accessor());
+        }
+
+        let source_permutation = permutation.map(|axes| self.rebase.invert_axes(axes));
+
+        if needs_transpose(&source_permutation) {
+            self.source
+                .transpose(source_permutation)
+                .map(|access| access.accessor())
+        } else {
+            Ok(self.source.accessor())
+        }
     }
 }
 
@@ -2162,7 +2393,6 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = Self;
-    type Transpose = Self;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Unary(Box::new(self))
@@ -2176,6 +2406,7 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseUnary::filled_at {:?}", axes);
         self.source.filled_at(txn, axes).await
     }
 
@@ -2183,6 +2414,10 @@ where
         self.source.filled_count(txn).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.source.is_empty(txn).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         let source = self.source.slice(bounds)?;
         Ok(SparseUnary {
@@ -2192,13 +2427,21 @@ where
         })
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let source = self.source.transpose(permutation)?;
-        Ok(SparseUnary {
-            source: source.accessor(),
-            dtype: self.dtype,
-            transform: self.transform,
-        })
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseUnary::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            let source = self.source.transpose(permutation)?;
+            let transpose = SparseUnary {
+                source: source.accessor(),
+                dtype: self.dtype,
+                transform: self.transform,
+            };
+
+            Ok(transpose.accessor())
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
