@@ -319,22 +319,23 @@ where
     D: Dir,
     T: Transaction<D>,
     A: SparseAccess<FD, FS, D, T>,
+    Self: TensorInstance,
+    <Self as TensorInstance>::Dense: TensorBooleanConst,
 {
     type Combine = SparseTensor<FD, FS, D, T, SparseConstCombinator<FD, FS, D, T, A>>;
+    type DenseCombine = <<Self as TensorInstance>::Dense as TensorBooleanConst>::DenseCombine;
 
     fn and_const(self, other: Number) -> TCResult<Self::Combine> {
         let access = SparseConstCombinator::new(self.accessor, other, Number::and);
         Ok(access.into())
     }
 
-    fn or_const(self, other: Number) -> TCResult<Self::Combine> {
-        let access = SparseConstCombinator::new(self.accessor, other, Number::or);
-        Ok(access.into())
+    fn or_const(self, other: Number) -> TCResult<Self::DenseCombine> {
+        self.into_dense().or_const(other)
     }
 
-    fn xor_const(self, other: Number) -> TCResult<Self::Combine> {
-        let access = SparseConstCombinator::new(self.accessor, other, Number::xor);
-        Ok(access.into())
+    fn xor_const(self, other: Number) -> TCResult<Self::DenseCombine> {
+        self.into_dense().xor_const(other)
     }
 }
 
@@ -946,11 +947,21 @@ where
     }
 }
 
-impl<FD, FS, D, T, A> TensorMathConst for SparseTensor<FD, FS, D, T, A> {
+impl<FD, FS, D, T, A> TensorMathConst for SparseTensor<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+    Self: TensorInstance,
+    <Self as TensorInstance>::Dense: TensorMathConst,
+{
     type Combine = SparseTensor<FD, FS, D, T, SparseConstCombinator<FD, FS, D, T, A>>;
+    type DenseCombine = <<Self as TensorInstance>::Dense as TensorMathConst>::DenseCombine;
 
-    fn add_const(self, other: Number) -> TCResult<Self::Combine> {
-        Ok(SparseConstCombinator::new(self.accessor, other, Number::add).into())
+    fn add_const(self, other: Number) -> TCResult<Self::DenseCombine> {
+        self.into_dense().add_const(other)
     }
 
     fn div_const(self, other: Number) -> TCResult<Self::Combine> {
@@ -986,8 +997,8 @@ impl<FD, FS, D, T, A> TensorMathConst for SparseTensor<FD, FS, D, T, A> {
         Ok(SparseConstCombinator::new(self.accessor, other, Number::pow).into())
     }
 
-    fn sub_const(self, other: Number) -> TCResult<Self::Combine> {
-        Ok(SparseConstCombinator::new(self.accessor, other, Number::sub).into())
+    fn sub_const(self, other: Number) -> TCResult<Self::DenseCombine> {
+        self.into_dense().sub_const(other)
     }
 }
 
@@ -1063,17 +1074,21 @@ where
     D::FileClass: From<TensorType>,
     A: SparseAccess<FD, FS, D, T>,
 {
-    type Broadcast = SparseTensor<FD, FS, D, T, SparseBroadcast<FD, FS, D, T, A>>;
+    type Broadcast = SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>;
     type Cast = SparseTensor<FD, FS, D, T, SparseCast<FD, FS, D, T, A>>;
     type Expand = SparseTensor<FD, FS, D, T, SparseExpand<FD, FS, D, T, A>>;
     type Flip = SparseTensor<FD, FS, D, T, SparseFlip<FD, FS, D, T, A>>;
     type Reshape = SparseTensor<FD, FS, D, T, SparseReshape<FD, FS, D, T, A>>;
     type Slice = SparseTensor<FD, FS, D, T, A::Slice>;
-    type Transpose = SparseTensor<FD, FS, D, T, A::Transpose>;
+    type Transpose = SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>;
 
     fn broadcast(self, shape: Shape) -> TCResult<Self::Broadcast> {
+        if self.shape() == &shape {
+            return Ok(self.into_inner().accessor().into());
+        }
+
         let accessor = SparseBroadcast::new(self.accessor, shape)?;
-        Ok(accessor.into())
+        Ok(accessor.accessor().into())
     }
 
     fn cast_into(self, dtype: NumberType) -> TCResult<Self::Cast> {
@@ -1102,6 +1117,7 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!("SparseTensor::transpose {:?}", permutation);
         let accessor = self.accessor.transpose(permutation)?;
         Ok(accessor.into())
     }
