@@ -8,7 +8,7 @@ use async_trait::async_trait;
 use destream::de;
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
-use log::debug;
+use log::{debug, trace};
 use safecast::AsType;
 
 use tc_btree::{BTreeType, Node};
@@ -23,8 +23,9 @@ use tcgeneric::{label, Id, Label, TCBoxTryStream, Tuple};
 
 use crate::dense::PER_BLOCK;
 use crate::stream::{sorted_coords, Read, ReadValueAt};
-use crate::transform;
-use crate::{AxisBounds, Bounds, Coord, Schema, Shape, TensorAccess, TensorType};
+use crate::{
+    needs_transpose, transform, AxisBounds, Bounds, Coord, Schema, Shape, TensorAccess, TensorType,
+};
 
 use super::access::SparseTranspose;
 use super::{SparseAccess, SparseAccessor, SparseStream, SparseTensor, SparseWrite};
@@ -106,21 +107,25 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
-    type Transpose = SparseTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Table(self)
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!("SparseTable::filled");
         let rows = self.table.rows(*txn.id()).await?;
-        let filled = rows.and_then(|row| future::ready(expect_row(row)));
+        let filled = rows
+            .map(|result| result.and_then(|row| expect_row(row)))
+            .inspect_ok(|(coord, value)| trace!("filled at {:?}: {}", coord, value));
+
         let filled: SparseStream = Box::pin(filled);
         Ok(filled)
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
         debug!("SparseTable::filled_at {:?}", axes);
+        debug_assert!(!axes.is_empty());
 
         self.shape().validate_axes(&axes)?;
 
@@ -134,6 +139,7 @@ where
         let shape = axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>();
         let coords = filled_at::<FD, FS, D, T, _>(&txn, axes, self.table).await?;
 
+        let coords = coords.inspect_ok(|coord| trace!("SparseTable filled at {:?}", coord));
         let coords = CoordBlocks::new(coords, shape.len(), PER_BLOCK);
 
         if sort {
@@ -148,6 +154,10 @@ where
         self.table.count(*txn.id()).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        self.table.is_empty(&txn).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         self.shape().validate_bounds(&bounds)?;
         let table_bounds = table_bounds(self.shape(), &bounds)?;
@@ -168,8 +178,13 @@ where
         }
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        SparseTranspose::new(self, permutation)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseTable::transpose {:?}", permutation);
+        if needs_transpose(&permutation) {
+            SparseTranspose::new(self, permutation).map(SparseAccess::accessor)
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -392,7 +407,6 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
-    type Transpose = SparseTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
         SparseAccessor::Slice(self)
@@ -412,6 +426,8 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseTableSlice::filled_at {:?}", axes);
+
         self.shape().validate_axes(&axes)?;
 
         if axes.is_empty() {
@@ -425,7 +441,8 @@ where
         let source_coords = filled_at::<FD, FS, D, T, _>(&txn, source_axes, self.table).await?;
         let coords = CoordBlocks::new(source_coords, self.source.ndim(), PER_BLOCK)
             .map_ok(move |coords| rebase.map_coords(coords))
-            .map_ok(move |coords| coords.get(&axes));
+            .map_ok(move |coords| coords.get(&axes))
+            .inspect_ok(|coords| trace!("SparseTableSlice {} coords to sort", coords.len()));
 
         let filled_at = sorted_coords::<FD, FS, D, T, _>(&txn, shape.into(), coords).await?;
         Ok(Box::pin(filled_at))
@@ -435,6 +452,11 @@ where
         self.table.count(*txn.id()).await
     }
 
+    async fn is_empty(self, txn: T) -> TCResult<bool> {
+        let mut rows = self.table.rows(*txn.id()).await?;
+        rows.try_next().map_ok(|row| row.is_none()).await
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseTableSlice::slice {}", bounds);
         self.shape().validate_bounds(&bounds)?;
@@ -449,8 +471,14 @@ where
         self.source.slice(source_bounds)
     }
 
-    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        SparseTranspose::new(self, permutation)
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<SparseAccessor<FD, FS, D, T>> {
+        debug!("SparseTableSlice::transpose {:?}", permutation);
+
+        if needs_transpose(&permutation) {
+            SparseTranspose::new(self, permutation).map(SparseAccess::accessor)
+        } else {
+            Ok(self.accessor())
+        }
     }
 }
 
@@ -539,9 +567,17 @@ where
     T::Selection: TableStream,
 {
     assert!(!axes.is_empty());
-    let coords = table.select(axes.into_iter().map(Id::from).collect())?;
+    let order: Vec<Id> = axes.into_iter().map(Id::from).collect();
+    debug!(
+        "sparse::table::filled_at {}",
+        order.iter().collect::<Tuple<&Id>>()
+    );
+
+    let coords = table.select(order)?;
     let coords = coords.rows(*txn.id()).await?;
-    Ok(coords.map(|r| r.and_then(expect_coord)))
+    Ok(coords
+        .map(|r| r.and_then(expect_coord))
+        .inspect_ok(|coord| trace!("sparse::table::filled_at coord {:?}", coord)))
 }
 
 fn table_bounds(shape: &Shape, bounds: &Bounds) -> TCResult<tc_table::Bounds> {
@@ -636,6 +672,7 @@ fn expect_coord(coord: Vec<Value>) -> TCResult<Coord> {
 fn expect_row(mut row: Vec<Value>) -> TCResult<(Coord, Number)> {
     if let Some(value) = row.pop() {
         let value = value.try_into()?;
+        debug_assert_ne!(value, 0.into());
         expect_coord(row).map(|coord| (coord, value))
     } else {
         Err(TCError::internal(ERR_CORRUPT))
