@@ -1,6 +1,8 @@
+use std::fmt;
+
 use afarray::{Array, ArrayExt, CoordUnique, Coords, Offsets};
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
-use log::debug;
+use log::{debug, trace};
 use safecast::AsType;
 
 use tc_btree::Node;
@@ -34,6 +36,7 @@ where
         .create_file_unique(txn_id, TensorType::Dense)
         .await?;
 
+    let coords = coords.inspect_ok(|block| trace!("coord block len is {}", block.len()));
     let offsets = sort_coords::<FD, FS, D, T, _>(file, txn_id, coords, shape.clone()).await?;
     debug!("sorted coords");
 
@@ -58,21 +61,20 @@ where
     FS: File<Node>,
     D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
-    A: TensorAccess + ReadValueAt<D, Txn = T> + Clone + 'a,
+    A: TensorAccess + ReadValueAt<D, Txn = T> + Clone + fmt::Display + 'a,
     C: Stream<Item = TCResult<Coords>> + Send + Unpin + 'a,
 {
-    debug!(
-        "sort values by coordinate for Tensor with shape {}",
-        source.shape()
-    );
+    debug!("sort values by coordinate for {}", source);
 
     let coords = sorted_coords::<FD, FS, D, T, C>(&txn, source.shape().clone(), coords).await?;
 
     let buffered = coords
         .map_ok(|coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
         .try_flatten()
+        .inspect_ok(|coord| trace!("read value at {:?} to sort", coord))
         .map_ok(move |coord| source.clone().read_value_at(txn.clone(), coord))
-        .try_buffered(num_cpus::get());
+        .try_buffered(num_cpus::get())
+        .inspect_ok(|(coord, value)| trace!("value at {:?} is {}", coord, value));
 
     Ok(buffered)
 }
@@ -92,7 +94,8 @@ where
 {
     debug!("sort coords for Tensor with shape {}", shape);
 
-    let blocks = coords_to_offsets(shape, coords).map_ok(|block| ArrayExt::from(block).into());
+    let blocks =
+        coords_to_offsets(shape, coords).map_ok(|block: Offsets| ArrayExt::from(block).into());
 
     let block_list =
         BlockListFile::from_blocks(file, txn_id, None, UIntType::U64.into(), Box::pin(blocks))
